#include <nemo/stdint.h>
#include <nemo/unixtypes.h>

#include <nemo/sig.h>
#include <nemo/strerr.h>
#include <nemo/error.h>
#include <nemo/fifo.h>
#include <nemo/open.h>
#include <nemo/lock.h>
#include <nemo/wait.h>
#include <nemo/coe.h>
#include <nemo/ndelay.h>
#include <nemo/env.h>
#include <nemo/iopause.h>
#include <nemo/tai.h>
#include <nemo/taia.h>
#include <nemo/exit.h>
#include <nemo/unix.h>
#include <nemo/fork.h>
#include <nemo/macro_unused.h>

#include "deepsleep.h"
#include "die.h"

#include <sys/stat.h>
#include <signal.h>

const char USAGE[] = "dir";

static const char WARNING[] = "supervise: warning: ";

static const char SUPERVISE_CONTROL[] = "supervise/control";
static const char SUPERVISE_LOCK[] = "supervise/lock";
static const char SUPERVISE_OK[] = "supervise/ok";
static const char FN_DOWN[] = "down";

static char *dir;
static int selfpipe[2];
static int fd_lock;
static int fd_control_write;
static int fd_control;
static int fd_ok;

static pid_t pid = 0;  /* 0 means down */
static unsigned int flag_paused;  /* defined if(pid) */
static unsigned int flag_exit = 0;
static unsigned int flag_want = 1;
static unsigned int flag_want_up = 1;

static byte_t status[18];

static void pid_change(void)
{
  struct taia now;
  pid_t u;

  taia_now(&now);
  taia_pack(&now, status);

  u = pid;

  status[12] = (byte_t)(u & 0xff);
  u >>= 8;

  status[13] = (byte_t)(u & 0xff);
  u >>= 8;

  status[14] = (byte_t)(u & 0xff);
  u >>= 8;

  status[15] = (byte_t)(u & 0xff);
}

static void announce(void)
{
  int fd;
  ssize_t r;

  status[16] = (byte_t)(pid ? flag_paused : 0);
  status[17] = (flag_want ? (flag_want_up ? 'u' : 'd') : 0);

  fd = open_trunc("supervise/status.new");
  if (fd < 0) {
    strerr_warn4sys(WARNING, "unable to open ", dir, "/supervise/status.new: ");
    return;
  }

  r = write(fd, status, (unsigned int)sizeof(status));
  if (r < 0) {
    strerr_warn4sys(WARNING, "unable to write ", dir, "/supervise/status.new");
    close(fd);
    return;
  }
  close(fd);
  if (r < (int)sizeof(status)) {
    strerr_warn4x(WARNING, "unable to write ", dir, "/supervise/status.new: partial write");
    return;
  }

  if (rename("supervise/status.new", "supervise/status") < 0) {
    strerr_warn4sys(WARNING, "unable to rename ", dir, "/supervise/status.new to status");
  }
}

static void trigger(int i __UNUSED__)
{
  write(selfpipe[1], "", 1);
}

static char runstr[] = "./run";
static char *run[2] = { runstr,  0 };

static void try_start(void)
{
  pid_t f;

  switch (f = fork()) {
    case -1:
      strerr_warn4sys(WARNING, "unable to fork for ", dir, ",  sleeping 60 seconds");
      deepsleep(60);
      trigger(0);
      return;
    case 0:
      sig_uncatch(sig_child);
      sig_unblock(sig_child);
      execve(*run, run, environ);
      die_exec2(dir, "run");
      break;
    default:
      break;
  }
  flag_paused = 0;
  pid = f;
  pid_change();
  announce();
  deepsleep(1);
}

static void doit(void)
{
  iopause_fd x[2];
  struct taia deadline;
  struct taia stamp;
  int wstat;
  int r;
  char ch;

  announce();

  for (;;) {
    if (flag_exit && !pid) return;

    sig_unblock(sig_child);

    x[0].fd = selfpipe[0];
    x[0].events = IOPAUSE_READ;
    x[1].fd = fd_control;
    x[1].events = IOPAUSE_READ;
    taia_now(&stamp);
    taia_uint(&deadline, 3600);
    taia_add(&deadline, &stamp, &deadline);
    iopause(x, 2, &deadline, &stamp);

    sig_block(sig_child);

    while (read(selfpipe[0], &ch, 1) == 1) {
      ;
    }

    for (;;) {
      r = wait_nohang(&wstat);
      if (!r) break;
      if ((r < 0) && (errno != error_intr)) break;
      if (r == pid) {
        pid = 0;
        pid_change();
        announce();
        if (flag_exit) return;
        if (flag_want && flag_want_up) {
          try_start();
        }
        break;
      }
    }

    if (read(fd_control, &ch, 1) == 1) {
      switch (ch) {
        case 'd':
          flag_want = 1;
          flag_want_up = 0;
          if (pid) {
            kill(pid, SIGTERM);
            kill(pid, SIGCONT);
            flag_paused = 0;
          }
          announce();
          break;
        case 'u':
          flag_want = 1;
          flag_want_up = 1;
          announce();
          if (!pid) {
            try_start();
          }
          break;
        case 'o':
          flag_want = 0;
          announce();
          if (!pid) {
            try_start();
          }
          break;
        case 'a':
          if (pid) {
            kill(pid, SIGALRM);
          }
          break;
        case 'h':
          if (pid) {
            kill(pid, SIGHUP);
          }
          break;
        case 'k':
          if (pid) {
            kill(pid, SIGKILL);
          }
          break;
        case 't':
          if (pid) {
            kill(pid, SIGTERM);
          }
          break;
        case 'i':
          if (pid) {
            kill(pid, SIGINT);
          }
          break;
        case 'p':
          flag_paused = 1;
          announce();
          if (pid) {
            kill(pid, SIGSTOP);
          }
          break;
        case 'c':
          flag_paused = 0;
          announce();
          if (pid) {
            kill(pid, SIGCONT);
          }
          break;
        case 'x':
          flag_exit = 1;
          announce();
          break;
        default:
          break;
      }
    }
  }
}

int main(int argc, char **argv)
{
  struct stat st;

  PROGRAM = argv[0];
  if (argc != 2) die_usage();
  dir = argv[1];

  if (pipe(selfpipe) < 0) die_create_pipe(dir);
  coe(selfpipe[0]);
  coe(selfpipe[1]);
  ndelay_on(selfpipe[0]);
  ndelay_on(selfpipe[1]);

  sig_block(sig_child);
  sig_catch(sig_child, trigger);

  if (chdir(dir) < 0) die_chdir(dir);

  if (stat(FN_DOWN, &st) < 0) {
    if (errno != error_noent) die_stat(dir, FN_DOWN);
  }
  else {
    flag_want_up = 0;
  }

  mkdir("supervise", 0700);
  fd_lock = open_append(SUPERVISE_LOCK);
  if ((fd_lock < 0) || (lock_exnb(fd_lock) < 0)) die_acquire(dir, SUPERVISE_LOCK);
  coe(fd_lock);

  fifo_make(SUPERVISE_CONTROL, 0600);
  fd_control = open_read(SUPERVISE_CONTROL);
  if (fd_control < 0) die_read2(dir, SUPERVISE_CONTROL);
  coe(fd_control);
  ndelay_on(fd_control); /* shouldn't be necessary */
  fd_control_write = open_write(SUPERVISE_CONTROL);
  if (fd_control_write < 0) die_write2(dir, SUPERVISE_CONTROL);
  coe(fd_control_write);

  pid_change();
  announce();

  fifo_make(SUPERVISE_OK, 0600);
  fd_ok = open_read(SUPERVISE_OK);
  if (fd_ok < 0) die_read2(dir, SUPERVISE_OK);
  coe(fd_ok);

  if (!flag_want || flag_want_up) {
    try_start();
  }

  doit();
  announce();
  _exit(0);
}
