#include <nemo/stdint.h>
#include <nemo/unixtypes.h>

#include <nemo/unix.h>
#include <nemo/direntry.h>
#include <nemo/strerr.h>
#include <nemo/error.h>
#include <nemo/wait.h>
#include <nemo/coe.h>
#include <nemo/fd.h>
#include <nemo/env.h>
#include <nemo/str.h>
#include <nemo/byte.h>
#include <nemo/pathexec.h>
#include <nemo/fork.h>

#include <sys/stat.h>

#include "die.h"

#define SERVICES 1000

static const char WARNING[] = "svscan: warning: ";

static struct {
  unsigned long dev;
  unsigned long ino;
  unsigned int flag_active;
  unsigned int flag_log;
  pid_t pid;	/* 0 if not running */
  pid_t pidlog;	/* 0 if not running */
  int pi[2];	/* defined if flag_log */
} x[SERVICES];

static unsigned int numx = 0;

static char fnlog[260];

static char supervise_str[] = "supervise";
static char log_str[] = "log";

static void start(char *fn)
{
  unsigned int fnlen;
  struct stat st;
  pid_t child;
  unsigned int i;
  char *args[3];

  if (fn[0] == '.') return;

  if (stat(fn, &st) < 0) {
    strerr_warn3sys(WARNING, "unable to stat ", fn);
    return;
  }

  if ((st.st_mode & S_IFMT) != S_IFDIR) return;

  for (i = 0; i < numx; ++i) {
    if (x[i].ino == st.st_ino) {
      if (x[i].dev == st.st_dev) {
        break;
      }
    }
  }

  if (i == numx) {
    if (numx >= SERVICES) {
      strerr_warn4x(WARNING, "unable to start ", fn, ": running too many services");
      return;
    }
    x[i].ino = st.st_ino;
    x[i].dev = (unsigned long)st.st_dev;
    x[i].pid = 0;
    x[i].pidlog = 0;
    x[i].flag_log = 0;

    fnlen = str_len(fn);
    if (fnlen <= 255) {
      byte_copy(fnlog, fnlen, fn);
      byte_copy(fnlog + fnlen, 5, "/log");
      if (stat(fnlog, &st) == 0) {
        x[i].flag_log = 1;
      }
      else {
        if (errno != error_noent) {
          strerr_warn4sys(WARNING, "unable to stat ", fn, "/log");
          return;
        }
      }
    }

    if (x[i].flag_log) {
      if (pipe(x[i].pi) < 0) {
        strerr_warn3sys(WARNING, "unable to create pipe for ", fn);
        return;
      }
      coe(x[i].pi[0]);
      coe(x[i].pi[1]);
    }
    ++numx;
  }

  x[i].flag_active = 1;

  if (!x[i].pid) {
    switch (child = fork()) {
      case -1:
        strerr_warn3sys(WARNING, "unable to fork for ", fn);
        return;
      case 0:  /* child */
        if (x[i].flag_log) {
          if (fd_move(1, x[i].pi[1]) < 0) {
            strerr_die3sys(111, WARNING, "unable to set up descriptors for ", fn);
          }
        }
        args[0] = supervise_str;
        args[1] = fn;
        args[2] = 0;
        pathexec_run(*args, args, environ);
        die_supervise(fn);
        break;
      default:
        x[i].pid = child;
    }
  }

  if (x[i].flag_log && !x[i].pidlog)
    switch (child = fork()) {
      case -1:
        strerr_warn4sys(WARNING, "unable to fork for ", fn, "/log");
        return;
      case 0:  /* child */
        if (fd_move(0, x[i].pi[0]) < 0) {
          strerr_die4sys(111, WARNING, "unable to set up descriptors for ", fn, "/log");
        }
        if (chdir(fn) < 0) die_chdir(fn);
        args[0] = supervise_str;
        args[1] = log_str;
        args[2] = 0;
        pathexec_run(*args, args, environ);
        die_supervise2(fn, "log");
        break;
      default:
        x[i].pidlog = child;
    }
}

static void direrror(void)
{
  strerr_warn2sys(WARNING, "unable to read directory");
}

static void doit(void)
{
  DIR *dir;
  direntry *d;
  unsigned int i;
  int r;
  int wstat;

  for (;;) {
    r = wait_nohang(&wstat);
    if (!r) break;
    if (r < 0) {
      if (errno == error_intr) continue; /* impossible */
      break;
    }

    for (i = 0; i < numx; ++i) {
      if (x[i].pid == r) {
        x[i].pid = 0;
        break;
      }
      if (x[i].pidlog == r) {
	x[i].pidlog = 0;
	break;
      }
    }
  }

  for (i = 0; i < numx; ++i) {
    x[i].flag_active = 0;
  }

  dir = opendir(".");
  if (!dir) {
    direrror();
    return;
  }
  for (;;) {
    errno = 0;
    d = readdir(dir);
    if (!d) break;
    start(d->d_name);
  }
  if (errno) {
    direrror();
    closedir(dir);
    return;
  }
  closedir(dir);

  i = 0;
  while (i < numx) {
    if (!x[i].flag_active && !x[i].pid && !x[i].pidlog) {
      if (x[i].flag_log) {
        close(x[i].pi[0]);
        close(x[i].pi[1]);
        x[i].flag_log = 0;
      }
      x[i] = x[--numx];
      continue;
    }
    ++i;
  }
}

int main(int argc, char **argv)
{
  PROGRAM = argv[0];
  if (argc > 1) {
    if (chdir(argv[1]) < 0) die_chdir(argv[1]);
  }

  for (;;) {
    doit();
    sleep(5);
  }
}
